using System;
using System.Linq;
using System.Collections.Generic;
using UnityEngine;
using System.Collections;
using TMPro;
using UnityEditor;
using static GlobalBehaviour;

/// <summary>
/// <see cref="Fact.Id"/>/ <c>MonoBehaviour</c> wrapper to be attached to <see cref="Fact.WorldRepresentation"/>
/// </summary>
//[DisallowMultipleComponent]
public abstract class FactObject : FactWrapper, ISerializationCallbackReceiver
{
    [SerializeField] protected List<TMP_Text> FactText;
    [SerializeField] protected List<string> StringLabelFormats;

    [NonSerialized]
    public Material[] Materials;
    protected Material[] MaterialsToChange;
    [NonSerialized]
    protected List<FactObject> AllChildren;

    #region Unity Serialization
    public Material Default;
    public Material Selected;
    public Material Hint;
    public Material Solution;

    void ISerializationCallbackReceiver.OnBeforeSerialize()
    {
        if (FactText == null || FactText.Count() == 0)
        {
            FactText = transform
                .GetComponentsInChildren<TMP_Text>(includeInactive: true)
                .ToList();
        }

        if (StringLabelFormats == null)
        {
            StringLabelFormats = new();

            int i = StringLabelFormats.Count();
            int end = FactText.Count();

            for (; i < end; i++)
                StringLabelFormats.Add("{" + (i + 1) + "}");
        }

        if (Materials != null)
        {
            Default = Materials[(int)FactMaterials.Default];
            Selected = Materials[(int)FactMaterials.Selected];
            Hint = Materials[(int)FactMaterials.Hint];
            Solution = Materials[(int)FactMaterials.Solution];
        }

        _OnBeforeSerialize();
    }
    protected abstract void _OnBeforeSerialize();

    void ISerializationCallbackReceiver.OnAfterDeserialize()
    {
        {
            Materials = new Material[4];
            Materials[(int)FactMaterials.Default] = Default;
            Materials[(int)FactMaterials.Selected] = Selected;
            Materials[(int)FactMaterials.Hint] = Hint;
            Materials[(int)FactMaterials.Solution] = Solution;
        }

        _OnAfterDeserialize();
    }
    protected abstract void _OnAfterDeserialize();
    #endregion

    protected override void _Awake()
    {
        base._Awake();

        AllChildren = transform.GetComponentsInChildren<FactObject>(includeInactive: true).ToList();
        AllChildren.Remove(this);

        FactText ??= new();
        StringLabelFormats ??= new();
    }

    public override void HighlightAs(FactMaterials tmpState)
    {
        CoroutineCascadeForMeAndChildrenAllMaterials(
            (fact_obj, material) => _Animation(fact_obj, material)
        );

        IEnumerator _Animation(FactObject fact_obj, Material material)
        {
            for (IEnumerator<float> clock = IEnumeratorExtensions.ClockForSeconds(HintAnimationDuration)
                ; clock.MoveNext();)
            {
                material.Lerp(fact_obj.Default,
                              fact_obj.Materials[(int)tmpState],
                              Mathf.PingPong(clock.Current, AnimationLerpDuration));

                yield return null;
            }

            // clean fix for UI: "material = fact_obj.Default" not working
            material.Lerp(fact_obj.Default,
                          fact_obj.Materials[(int)tmpState],
                          0);
        }
    }

    protected override void FactUpdated()
    {
        Awake();

        ReLabel();

        foreach (FactObject childObject in AllChildren)
        {
            childObject._URI = _URI;
            childObject._Fact = _Fact;
            childObject.ReLabel();
        }
    }

    private void ReLabel()
    {
        string[] mother_child_labels =
            new[] { URI }.ShallowCloneAppend(Fact.DependentFactIds)
            .Select(fid =>
                FactOrganizer.AllFacts.TryGetValue(fid, out Fact fact)
                ? fact.Label
                : "N/A")
            .ToArray();

        for (int i = 0; i < Math.Min(FactText.Count(), StringLabelFormats.Count()); i++)
        {
            try
            {
                FactText[i].text = string.Format(StringLabelFormats[i], mother_child_labels);
            }
            catch (Exception ex)
            {
                Debug.LogError(ex);
                FactText[i].text = "Err";
            }
        }

        _CustomReLabel();
    }
    protected virtual void _CustomReLabel() { }

    public void CascadeForMeAndChildren(Action<FactObject> func)
    {
        func(this);
        foreach (FactObject fo in AllChildren)
            func(fo);
    }

    public void ForAllMaterials(Action<Material> func)
    {
        foreach (Material mat in MaterialsToChange)
            func(mat);
    }

    public void CoroutineCascadeForMeAndChildrenAllMaterials(Func<FactObject, Material, IEnumerator> func)
    {
        this.StopAllCoroutines();

        CascadeForMeAndChildren((FactObject fo) =>
            fo.ForAllMaterials((Material mat) =>
                this.StartCoroutine(func(fo, mat))
        ));
    }
}
